@@ -15,16 +15,24 @@
 
 """Worry-free YAML configuration files.
 """
-from __future__ import division, absolute_import, print_function
+from __future__ import annotations
 
+__all__ = [
+    'CONFIG_FILENAME', 'DEFAULT_FILENAME', 'ROOT_NAME', 'REDACTED_TOMBSTONE',
+    'ConfigView', 'RootView', 'Subview', 'Configuration', 'LazyConfig'
+]
+
+import errno
 import os
+from pathlib import Path
+from typing import Any, Iterable, Sequence, TypeVar
 import yaml
 from collections import OrderedDict
 
 from . import util
 from . import templates
 from . import yaml_util
-from .sources import ConfigSource, YamlSource
+from .sources import ConfigSource, EnvSource, YamlSource
 from .exceptions import ConfigTypeError, NotFoundError, ConfigError
 
 CONFIG_FILENAME = 'config.yaml'
@@ -33,11 +41,12 @@ ROOT_NAME = 'root'
 
 REDACTED_TOMBSTONE = 'REDACTED'
 
+R = TypeVar('R')
 
 # Views and sources.
 
 
-class ConfigView(object):
+class ConfigView():
     """A configuration "view" is a query into a program's configuration
     data. A view represents a hypothetical location in the configuration
     tree; to extract the data from the location, a client typically
@@ -69,7 +78,7 @@ class ConfigView(object):
         try:
             return util.iter_first(pairs)
         except ValueError:
-            raise NotFoundError(u"{0} not found".format(self.name))
+            raise NotFoundError("{0} not found".format(self.name))
 
     def exists(self):
         """Determine whether the view has a setting in any source.
@@ -120,7 +129,7 @@ class ConfigView(object):
             except ConfigTypeError:
                 item, _ = self.first()
                 raise ConfigTypeError(
-                    u'{0} must be a dictionary or a list, not {1}'.format(
+                    '{0} must be a dictionary or a list, not {1}'.format(
                         self.name, type(item).__name__
                     )
                 )
@@ -138,69 +147,6 @@ class ConfigView(object):
     def __contains__(self, key):
         return self[key].exists()
 
-    @classmethod
-    def _build_namespace_dict(cls, obj, dots=False):
-        """Recursively replaces all argparse.Namespace and optparse.Values
-        with dicts and drops any keys with None values.
-
-        Additionally, if dots is True, will expand any dot delimited
-        keys.
-
-        :param obj: Namespace, Values, or dict to iterate over. Other
-            values will simply be returned.
-        :type obj: argparse.Namespace or optparse.Values or dict or *
-        :param dots: If True, any properties on obj that contain dots (.)
-            will be broken down into child dictionaries.
-        :return: A new dictionary or the value passed if obj was not a
-            dict, Namespace, or Values.
-        :rtype: dict or *
-        """
-        # We expect our root object to be a dict, but it may come in as
-        # a namespace
-        obj = util.namespace_to_dict(obj)
-        # We only deal with dictionaries
-        if not isinstance(obj, dict):
-            return obj
-
-        # Get keys iterator
-        keys = obj.keys() if util.PY3 else obj.iterkeys()
-        if dots:
-            # Dots needs sorted keys to prevent parents from
-            # clobbering children
-            keys = sorted(list(keys))
-
-        output = {}
-        for key in keys:
-            value = obj[key]
-            if value is None:  # Avoid unset options.
-                continue
-
-            save_to = output
-            result = cls._build_namespace_dict(value, dots)
-            if dots:
-                # Split keys by dots as this signifies nesting
-                split = key.split('.')
-                if len(split) > 1:
-                    # The last index will be the key we assign result to
-                    key = split.pop()
-                    # Build the dict tree if needed and change where
-                    # we're saving to
-                    for child_key in split:
-                        if child_key in save_to and \
-                                isinstance(save_to[child_key], dict):
-                            save_to = save_to[child_key]
-                        else:
-                            # Clobber or create
-                            save_to[child_key] = {}
-                            save_to = save_to[child_key]
-
-            # Save
-            if key in save_to:
-                save_to[key].update(result)
-            else:
-                save_to[key] = result
-        return output
-
     def set_args(self, namespace, dots=False):
         """Overlay parsed command-line arguments, generated by a library
         like argparse or optparse, onto this view's value.
@@ -217,7 +163,7 @@ class ConfigView(object):
             {'foo': {'bar': 'car'}}
         :type dots: bool
         """
-        self.set(self._build_namespace_dict(namespace, dots))
+        self.set(util.build_dict(namespace, sep='.' if dots else ''))
 
     # Magical conversions. These special methods make it possible to use
     # View objects somewhat transparently in certain circumstances. For
@@ -227,23 +173,10 @@ class ConfigView(object):
     def __str__(self):
         """Get the value for this view as a bytestring.
         """
-        if util.PY3:
-            return self.__unicode__()
-        else:
-            return bytes(self.get())
-
-    def __unicode__(self):
-        """Get the value for this view as a Unicode string.
-        """
-        return util.STRING(self.get())
-
-    def __nonzero__(self):
-        """Gets the value for this view as a boolean. (Python 2 only.)
-        """
-        return self.__bool__()
+        return str(self.get())
 
     def __bool__(self):
-        """Gets the value for this view as a boolean. (Python 3 only.)
+        """Gets the value for this view as a bool.
         """
         return bool(self.get())
 
@@ -265,7 +198,7 @@ class ConfigView(object):
                 cur_keys = dic.keys()
             except AttributeError:
                 raise ConfigTypeError(
-                    u'{0} must be a dict, not {1}'.format(
+                    '{0} must be a dict, not {1}'.format(
                         self.name, type(dic).__name__
                     )
                 )
@@ -305,7 +238,7 @@ class ConfigView(object):
             return
         if not isinstance(collection, (list, tuple)):
             raise ConfigTypeError(
-                u'{0} must be a list, not {1}'.format(
+                '{0} must be a list, not {1}'.format(
                     self.name, type(collection).__name__
                 )
             )
@@ -326,7 +259,7 @@ class ConfigView(object):
                 it = iter(collection)
             except TypeError:
                 raise ConfigTypeError(
-                    u'{0} must be an iterable, not {1}'.format(
+                    '{0} must be an iterable, not {1}'.format(
                         self.name, type(collection).__name__
                     )
                 )
@@ -354,7 +287,7 @@ class ConfigView(object):
                     od[key] = view.get()
         return od
 
-    def get(self, template=templates.REQUIRED):
+    def get(self, template=templates.REQUIRED) -> Any:
         """Retrieve the value for this view according to the template.
 
         The `template` against which the values are checked can be
@@ -371,47 +304,50 @@ class ConfigView(object):
 
     # Shortcuts for common templates.
 
-    def as_filename(self):
+    def as_filename(self) -> str:
         """Get the value as a path. Equivalent to `get(Filename())`.
         """
         return self.get(templates.Filename())
 
-    def as_path(self):
+    def as_path(self) -> Path:
         """Get the value as a `pathlib.Path` object. Equivalent to `get(Path())`.
         """
         return self.get(templates.Path())
 
-    def as_choice(self, choices):
+    def as_choice(self, choices: Iterable[R]) -> R:
         """Get the value from a list of choices. Equivalent to
         `get(Choice(choices))`.
+
+        Sequences, dictionaries and :class:`Enum` types are supported,
+        see :class:`confuse.templates.Choice` for more details.
         """
         return self.get(templates.Choice(choices))
 
-    def as_number(self):
+    def as_number(self) -> int | float:
         """Get the value as any number type: int or float. Equivalent to
         `get(Number())`.
         """
         return self.get(templates.Number())
 
-    def as_str_seq(self, split=True):
+    def as_str_seq(self, split=True) -> Sequence[str]:
         """Get the value as a sequence of strings. Equivalent to
         `get(StrSeq(split=split))`.
         """
         return self.get(templates.StrSeq(split=split))
 
-    def as_pairs(self, default_value=None):
+    def as_pairs(self, default_value=None) -> Sequence[tuple[str, str]]:
         """Get the value as a sequence of pairs of two strings. Equivalent to
         `get(Pairs(default_value=default_value))`.
         """
         return self.get(templates.Pairs(default_value=default_value))
 
-    def as_str(self):
+    def as_str(self) -> str:
         """Get the value as a (Unicode) string. Equivalent to
         `get(unicode)` on Python 2 and `get(str)` on Python 3.
         """
         return self.get(templates.String())
 
-    def as_str_expanded(self):
+    def as_str_expanded(self) -> str:
         """Get the value as a (Unicode) string, with env vars
         expanded by `os.path.expandvars()`.
         """
@@ -455,8 +391,8 @@ class RootView(ConfigView):
         self.name = ROOT_NAME
         self.redactions = set()
 
-    def add(self, obj):
-        self.sources.append(ConfigSource.of(obj))
+    def add(self, value):
+        self.sources.append(ConfigSource.of(value=value))
 
     def set(self, value):
         self.sources.insert(0, ConfigSource.of(value))
@@ -500,10 +436,10 @@ class Subview(ConfigView):
             if not isinstance(self.key, int):
                 self.name += '.'
         if isinstance(self.key, int):
-            self.name += u'#{0}'.format(self.key)
+            self.name += '#{0}'.format(self.key)
         elif isinstance(self.key, bytes):
             self.name += self.key.decode('utf-8')
-        elif isinstance(self.key, util.STRING):
+        elif isinstance(self.key, str):
             self.name += self.key
         else:
             self.name += repr(self.key)
@@ -521,7 +457,7 @@ class Subview(ConfigView):
             except TypeError:
                 # Not subscriptable.
                 raise ConfigTypeError(
-                    u"{0} must be a collection, not {1}".format(
+                    "{0} must be a collection, not {1}".format(
                         self.parent.name, type(collection).__name__
                     )
                 )
@@ -559,7 +495,7 @@ class Configuration(RootView):
         object at module load time and then call the `read` method
         later. Specify the Loader class as `loader`.
         """
-        super(Configuration, self).__init__([])
+        super().__init__([])
         self.appname = appname
         self.modname = modname
         self.loader = loader
@@ -629,7 +565,7 @@ class Configuration(RootView):
             appdir = os.environ[self._env_var]
             appdir = os.path.abspath(os.path.expanduser(appdir))
             if os.path.isfile(appdir):
-                raise ConfigError(u'{0} must be a directory'.format(
+                raise ConfigError('{0} must be a directory'.format(
                     self._env_var
                 ))
 
@@ -645,8 +581,12 @@ class Configuration(RootView):
                 appdir = os.path.join(configdirs[0], self.appname)
 
         # Ensure that the directory exists.
-        if not os.path.isdir(appdir):
+        try:
             os.makedirs(appdir)
+        except OSError as e:
+            if e.errno != errno.EEXIST:
+                raise
+
         return appdir
 
     def set_file(self, filename, base_for_paths=False):
@@ -662,6 +602,28 @@ class Configuration(RootView):
         self.set(YamlSource(filename, base_for_paths=base_for_paths,
                             loader=self.loader))
 
+    def set_env(self, prefix=None, sep='__'):
+        """Create a configuration overlay at the highest priority from
+        environment variables.
+
+        After prefix matching and removal, environment variable names will be
+        converted to lowercase for use as keys within the configuration. If
+        there are nested keys, list-like dicts (ie, `{0: 'a', 1: 'b'}`) will
+        be converted into corresponding lists (ie, `['a', 'b']`). The values
+        of all environment variables will be parsed as YAML scalars using the
+        `self.loader` Loader class to ensure type conversion is consistent
+        with YAML file sources. Use the `EnvSource` class directly to load
+        environment variables using non-default behavior and to enable full
+        YAML parsing of values.
+
+        :param prefix: The prefix to identify the environment variables to use.
+            Defaults to uppercased `self.appname` followed by an underscore.
+        :param sep: Separator within variable names to define nested keys.
+        """
+        if prefix is None:
+            prefix = '{0}_'.format(self.appname.upper())
+        self.set(EnvSource(prefix, sep=sep, loader=self.loader))
+
     def dump(self, full=True, redact=False):
         """Dump the Configuration object to a YAML file.
 
@@ -719,14 +681,14 @@ class LazyConfig(Configuration):
     the module level.
     """
     def __init__(self, appname, modname=None):
-        super(LazyConfig, self).__init__(appname, modname, False)
+        super().__init__(appname, modname, False)
         self._materialized = False  # Have we read the files yet?
         self._lazy_prefix = []  # Pre-materialization calls to set().
         self._lazy_suffix = []  # Calls to add().
 
     def read(self, user=True, defaults=True):
         self._materialized = True
-        super(LazyConfig, self).read(user, defaults)
+        super().read(user, defaults)
 
     def resolve(self):
         if not self._materialized:
@@ -734,17 +696,17 @@ class LazyConfig(Configuration):
             self.read()
             self.sources += self._lazy_suffix
             self.sources[:0] = self._lazy_prefix
-        return super(LazyConfig, self).resolve()
+        return super().resolve()
 
     def add(self, value):
-        super(LazyConfig, self).add(value)
+        super().add(value)
         if not self._materialized:
             # Buffer additions to end.
             self._lazy_suffix += self.sources
             del self.sources[:]
 
     def set(self, value):
-        super(LazyConfig, self).set(value)
+        super().set(value)
         if not self._materialized:
             # Buffer additions to beginning.
             self._lazy_prefix[:0] = self.sources
@@ -752,7 +714,7 @@ class LazyConfig(Configuration):
 
     def clear(self):
         """Remove all sources from this configuration."""
-        super(LazyConfig, self).clear()
+        super().clear()
         self._lazy_suffix = []
         self._lazy_prefix = []
 
